use std::collections::HashMap;
use std::marker::PhantomData;

use super::ast_node::IntoIAstNode;
use crate::error::CustomError;
use organic::types::Element;
use organic::types::Object;

use super::ast_node::IAstNode;

type IdCounter = u16;

pub(crate) struct Registry<'orig, 'parse> {
    id_counter: IdCounter,
    targets: HashMap<&'parse str, String>,
    footnote_ids: Vec<(Option<&'parse str>, Vec<IAstNode>)>,
    _phantom: PhantomData<&'orig ()>,
}

impl<'orig, 'parse> Registry<'orig, 'parse> {
    pub(crate) fn new() -> Registry<'orig, 'parse> {
        Registry {
            id_counter: 0,
            targets: HashMap::new(),
            footnote_ids: Vec::new(),
            _phantom: PhantomData,
        }
    }

    pub(crate) fn get_target<'reg>(&'reg mut self, body: &'parse str) -> &'reg String {
        self.targets.entry(body).or_insert_with(|| {
            self.id_counter += 1;
            format!("target_{}", self.id_counter)
        })
    }

    pub(crate) fn get_footnote_ids(&self) -> impl Iterator<Item = (usize, &Vec<IAstNode>)> {
        self.footnote_ids
            .iter()
            .map(|(_label, definition)| definition)
            .enumerate()
    }

    /// Get a 0-indexed ID for a footnote.
    ///
    /// This needs to be incremented to be 1-indexed for render.
    pub(crate) async fn get_footnote_reference_id<'reg>(
        &'reg mut self,
        label: Option<&'parse str>,
        definition: &'orig Vec<Object<'parse>>,
    ) -> Result<usize, CustomError> {
        if let None = label {
            // If it has no label then it must always get a new ID.
            let contents = convert_reference_contents(self, definition).await?;
            self.footnote_ids.push((None, contents));
            return Ok(self.footnote_ids.len() - 1);
        }

        if let Some(existing_id) = self
            .footnote_ids
            .iter()
            .position(|(id, _definition)| *id == label)
        {
            if !definition.is_empty() {
                let contents = convert_reference_contents(self, definition).await?;
                let entry = self
                    .footnote_ids
                    .get_mut(existing_id)
                    .expect("If-statement proves this to be Some.");
                entry.1 = contents;
            }
            Ok(existing_id)
        } else {
            let contents = convert_reference_contents(self, definition).await?;
            self.footnote_ids.push((label, contents));
            Ok(self.footnote_ids.len() - 1)
        }
    }

    /// Update the definition to a footnote but do not mark it as referenced.
    pub(crate) async fn register_footnote_definition<'reg>(
        &'reg mut self,
        label: &'parse str,
        definition: &'orig Vec<Element<'parse>>,
    ) -> Result<(), CustomError> {
        let contents = convert_definition_contents(self, definition).await?;
        if let Some((_existing_id, existing_definition)) = self
            .footnote_ids
            .iter_mut()
            .find(|(id, _definition)| *id == Some(label))
        {
            *existing_definition = contents;
        }
        Ok(())
    }
}

async fn convert_reference_contents<'reg, 'orig, 'parse>(
    registry: &'reg mut Registry<'orig, 'parse>,
    contents: &'orig Vec<Object<'parse>>,
) -> Result<Vec<IAstNode>, CustomError> {
    let contents = {
        let mut ret = Vec::new();
        for obj in contents.iter() {
            ret.push(obj.into_ast_node(registry).await?);
        }
        ret
    };

    Ok(contents)
}

async fn convert_definition_contents<'reg, 'orig, 'parse>(
    registry: &'reg mut Registry<'orig, 'parse>,
    contents: &'orig Vec<Element<'parse>>,
) -> Result<Vec<IAstNode>, CustomError> {
    let contents = {
        let mut ret = Vec::new();
        for obj in contents.iter() {
            ret.push(obj.into_ast_node(registry).await?);
        }
        ret
    };

    Ok(contents)
}
